/*
  ==============================================================================

   This file is part of the JUCE tutorials.
   Copyright (c) 2020 - Raw Material Software Limited

   The code included in this file is provided under the terms of the ISC license
   http://www.isc.org/downloads/software-support-policy/isc-license. Permission
   To use, copy, modify, and/or distribute this software for any purpose with or
   without fee is hereby granted provided that the above copyright notice and
   this permission notice appear in all copies.

   THE SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY, AND ALL WARRANTIES,
   WHETHER EXPRESSED OR IMPLIED, INCLUDING MERCHANTABILITY AND FITNESS FOR
   PURPOSE, ARE DISCLAIMED.

  ==============================================================================
*/

/*******************************************************************************
 The block below describes the properties of this PIP. A PIP is a short snippet
 of code that can be read by the Projucer and used to generate a JUCE project.

 BEGIN_JUCE_PIP_METADATA

 name:             SIMDRegisterTutorial
 version:          2.0.0
 vendor:           JUCE
 website:          http://juce.com
 description:      Explores SIMD register features.

 dependencies:     juce_audio_basics, juce_audio_devices, juce_audio_formats,
                   juce_audio_processors, juce_audio_utils, juce_core,
                   juce_data_structures, juce_dsp, juce_events, juce_graphics,
                   juce_gui_basics, juce_gui_extra
 exporters:        xcode_mac, vs2019, linux_make

 type:             Component
 mainClass:        AudioFileReaderComponent

 useLocalCopy:     1

 END_JUCE_PIP_METADATA

*******************************************************************************/


#pragma once

namespace dsp = juce::dsp;

//==============================================================================
struct DSPParameterBase    : public juce::ChangeBroadcaster
{
    DSPParameterBase (const juce::String& labelName) : name (labelName) {}
    virtual ~DSPParameterBase() {}

    virtual juce::Component* getComponent() = 0;

    virtual int getPreferredHeight()  = 0;
    virtual int getPreferredWidth()   = 0;

    juce::String name;

    JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (DSPParameterBase)
};

//==============================================================================
struct SliderParameter   : public DSPParameterBase
{
    SliderParameter (juce::Range<double> range, double skew, double initialValue,
                     const juce::String& labelName, const juce::String& suffix = {})
        : DSPParameterBase (labelName)
    {
        slider.setRange (range.getStart(), range.getEnd(), 0.01);
        slider.setSkewFactor (skew);
        slider.setValue (initialValue);

        if (suffix.isNotEmpty())
            slider.setTextValueSuffix (suffix);

        slider.onValueChange = [this] { sendChangeMessage(); };
    }

    juce::Component* getComponent() override    { return &slider; }

    int getPreferredHeight() override     { return 40; }
    int getPreferredWidth()  override     { return 500; }

    double getCurrentValue() const        { return slider.getValue(); }

private:
    juce::Slider slider;
};

//==============================================================================
struct ChoiceParameter    : public DSPParameterBase
{
    ChoiceParameter (const juce::StringArray& options, int initialId, const juce::String& labelName)
        : DSPParameterBase (labelName)
    {
        parameterBox.addItemList (options, 1);
        parameterBox.onChange = [this] { sendChangeMessage(); };

        parameterBox.setSelectedId (initialId);
    }

    juce::Component* getComponent() override    { return &parameterBox; }

    int getPreferredHeight() override     { return 25; }
    int getPreferredWidth()  override     { return 250; }

    int getCurrentSelectedID() const      { return parameterBox.getSelectedId(); }

private:
    juce::ComboBox parameterBox;
};

//! [SIMDTutorialFilter prefix]
#if JUCE_USE_SIMD

//==============================================================================
template <typename T>
static T* toBasePointer (dsp::SIMDRegister<T>* r) noexcept
{
    return reinterpret_cast<T*> (r);
}

constexpr auto registerSize = dsp::SIMDRegister<float>::size();

struct SIMDTutorialFilter
{
//! [SIMDTutorialFilter prefix]
//! [SIMDTutorialFilter prepare]
    void prepare (const dsp::ProcessSpec& spec)
    {
        sampleRate = spec.sampleRate;   // [4]

        iirCoefficients = dsp::IIR::Coefficients<float>::makeLowPass (sampleRate, 440.0f);
        iir.reset (new dsp::IIR::Filter<dsp::SIMDRegister<float>> (iirCoefficients)); // [5]

        interleaved = dsp::AudioBlock<dsp::SIMDRegister<float>> (interleavedBlockData, 1, spec.maximumBlockSize);
        zero        = dsp::AudioBlock<float> (zeroData, dsp::SIMDRegister<float>::size(), spec.maximumBlockSize); // [6]

        zero.clear();

        auto monoSpec = spec;
        monoSpec.numChannels = 1;
        iir->prepare (monoSpec);    // [7]
    }
//! [SIMDTutorialFilter prepare]

    template <typename SampleType>
    auto prepareChannelPointers (const dsp::AudioBlock<SampleType>& block)
    {
        std::array<SampleType*, registerSize> result {};

        for (size_t ch = 0; ch < result.size(); ++ch)
            result[ch] = (ch < block.getNumChannels() ? block.getChannelPointer (ch) : zero.getChannelPointer (ch));

        return result;
    }

//! [SIMDTutorialFilter process]
    void process (const dsp::ProcessContextReplacing<float>& context)
    {
        jassert (context.getInputBlock().getNumSamples()  == context.getOutputBlock().getNumSamples());
        jassert (context.getInputBlock().getNumChannels() == context.getOutputBlock().getNumChannels());

        const auto& input  = context.getInputBlock(); // [9]
        const auto numSamples = (int) input.getNumSamples();

        auto inChannels = prepareChannelPointers (input); // [10]

        using Format = juce::AudioData::Format<juce::AudioData::Float32, juce::AudioData::NativeEndian>;

        juce::AudioData::interleaveSamples (juce::AudioData::NonInterleavedSource<Format> { inChannels.data(),                                 registerSize, },
                                            juce::AudioData::InterleavedDest<Format>      { toBasePointer (interleaved.getChannelPointer (0)), registerSize },
                                            numSamples); // [11]

        iir->process (dsp::ProcessContextReplacing<dsp::SIMDRegister<float>> (interleaved)); // [12]

        auto outChannels = prepareChannelPointers (context.getOutputBlock()); // [13]

        juce::AudioData::deinterleaveSamples (juce::AudioData::InterleavedSource<Format>  { toBasePointer (interleaved.getChannelPointer (0)), registerSize },
                                              juce::AudioData::NonInterleavedDest<Format> { outChannels.data(),                                registerSize },
                                              numSamples); // [14]
    }
//! [SIMDTutorialFilter process]

    void reset()
    {
        iir.reset();
    }

//! [SIMDTutorialFilter updateParameters]
    void updateParameters()
    {
        if (sampleRate != 0.0)
        {
            auto cutoff = static_cast<float> (cutoffParam.getCurrentValue());
            auto qVal   = static_cast<float> (qParam.getCurrentValue());

            switch (typeParam.getCurrentSelectedID())
            {
                case 1:   *iirCoefficients = *dsp::IIR::Coefficients<float>::makeLowPass  (sampleRate, cutoff, qVal); break;
                case 2:   *iirCoefficients = *dsp::IIR::Coefficients<float>::makeHighPass (sampleRate, cutoff, qVal); break;
                case 3:   *iirCoefficients = *dsp::IIR::Coefficients<float>::makeBandPass (sampleRate, cutoff, qVal); break;
                default:  break;
            }
        }
    }
//! [SIMDTutorialFilter updateParameters]

    //==============================================================================
//! [SIMDTutorialFilter members]
    dsp::IIR::Coefficients<float>::Ptr iirCoefficients;                 // [1]
    std::unique_ptr<dsp::IIR::Filter<dsp::SIMDRegister<float>>> iir;

    dsp::AudioBlock<dsp::SIMDRegister<float>> interleaved;              // [2]
    dsp::AudioBlock<float> zero;

    juce::HeapBlock<char> interleavedBlockData, zeroData;               // [3]

    ChoiceParameter typeParam { { "Low-pass", "High-pass", "Band-pass" }, 1, "Type" };
    SliderParameter cutoffParam { { 20.0, 20000.0 }, 0.5, 440.0f, "Cutoff", "Hz" };
    SliderParameter qParam { { 0.3, 20.0 }, 0.5, 0.7, "Q" };

    std::vector<DSPParameterBase*> parameters { &typeParam, &cutoffParam, &qParam };
    double sampleRate = 0.0;
//! [SIMDTutorialFilter members]
//! [SIMDTutorialFilter suffix]
};

#endif
//! [SIMDTutorialFilter suffix]

//==============================================================================
struct DSPProcessor  : public juce::AudioSource,
                       public dsp::ProcessorWrapper<SIMDTutorialFilter>,
                       private juce::ChangeListener
{
    DSPProcessor (AudioSource& input)
        : inputSource (&input)
    {
        for (auto* p : getParameters())
            p->addChangeListener (this);
    }

    void prepareToPlay (int blockSize, double sampleRate) override
    {
        inputSource->prepareToPlay (blockSize, sampleRate);
        this->prepare ({ sampleRate, (juce::uint32) blockSize, 2 });
    }

    void releaseResources() override
    {
        inputSource->releaseResources();
    }

    void getNextAudioBlock (const juce::AudioSourceChannelInfo& bufferToFill) override
    {
        jassert (bufferToFill.buffer != nullptr);

        inputSource->getNextAudioBlock (bufferToFill);

        dsp::AudioBlock<float> block (*bufferToFill.buffer,
                                      (size_t) bufferToFill.startSample);

        juce::ScopedLock audioLock (audioCallbackLock);
        this->process (dsp::ProcessContextReplacing<float> (block));
    }

    const std::vector<DSPParameterBase*>& getParameters()
    {
        return this->processor.parameters;
    }

    void changeListenerCallback (juce::ChangeBroadcaster*) override
    {
        juce::ScopedLock audioLock (audioCallbackLock);
        static_cast<SIMDTutorialFilter&> (this->processor).updateParameters();
    }

    juce::CriticalSection audioCallbackLock;

    AudioSource* inputSource;
};

//==============================================================================
class AudioThumbnailComponent    : public juce::Component,
                                   public juce::FileDragAndDropTarget,
                                   public juce::ChangeBroadcaster,
                                   private juce::ChangeListener,
                                   private juce::Timer
{
public:
    AudioThumbnailComponent (juce::AudioDeviceManager& adm, juce::AudioFormatManager& afm)
        : audioDeviceManager (adm),
          thumbnailCache (5),
          thumbnail (128, afm, thumbnailCache)
    {
        thumbnail.addChangeListener (this);
    }

    ~AudioThumbnailComponent() override
    {
        thumbnail.removeChangeListener (this);
    }

    void paint (juce::Graphics& g) override
    {
        g.fillAll (juce::Colour (0xff495358));

        g.setColour (juce::Colours::white);

        if (thumbnail.getTotalLength() > 0.0)
        {
            thumbnail.drawChannels (g, getLocalBounds().reduced (2),
                                    0.0, thumbnail.getTotalLength(), 1.0f);

            g.setColour (juce::Colours::black);
            g.fillRect (static_cast<float> (currentPosition * getWidth()), 0.0f,
                        1.0f, static_cast<float> (getHeight()));
        }
        else
        {
            g.drawFittedText ("No audio file loaded.\nDrop a file here or click the \"Load File...\" button.", getLocalBounds(),
                              juce::Justification::centred, 2);
        }
    }

    bool isInterestedInFileDrag (const juce::StringArray&) override          { return true; }
    void filesDropped (const juce::StringArray& files, int, int) override    { loadFile (juce::File (files[0]), true); }

    void setCurrentFile (const juce::File& f)
    {
        if (currentFile == f)
            return;

        loadFile (f);
    }

    juce::File getCurrentFile()    { return currentFile; }

    void setTransportSource (juce::AudioTransportSource* newSource)
    {
        transportSource = newSource;

        struct ResetCallback  : public juce::CallbackMessage
        {
            ResetCallback (AudioThumbnailComponent& o) : owner (o) {}
            void messageCallback() override    { owner.reset(); }

            AudioThumbnailComponent& owner;
        };

        (new ResetCallback (*this))->post();
    }

private:
    juce::AudioDeviceManager& audioDeviceManager;
    juce::AudioThumbnailCache thumbnailCache;
    juce::AudioThumbnail thumbnail;
    juce::AudioTransportSource* transportSource = nullptr;

    juce::File currentFile;
    double currentPosition = 0.0;

    //==============================================================================
    void changeListenerCallback (ChangeBroadcaster*) override    { repaint(); }

    void reset()
    {
        currentPosition = 0.0;
        repaint();

        if (transportSource == nullptr)
            stopTimer();
        else
            startTimerHz (25);
    }

    void loadFile (const juce::File& f, bool notify = false)
    {
        if (currentFile == f || ! f.existsAsFile())
            return;

        currentFile = f;
        thumbnail.setSource (new juce::FileInputSource (f));

        if (notify)
            sendChangeMessage();
    }

    void timerCallback() override
    {
        if (transportSource != nullptr)
        {
            currentPosition = transportSource->getCurrentPosition() / thumbnail.getTotalLength();
            repaint();
        }
    }

    void mouseDrag (const juce::MouseEvent& e) override
    {
        if (transportSource != nullptr)
        {
            const juce::ScopedLock sl (audioDeviceManager.getAudioCallbackLock());

            transportSource->setPosition ((juce::jmax (static_cast<double> (e.x), 0.0) / getWidth())
                                            * thumbnail.getTotalLength());
        }
    }
};

//==============================================================================
class DSPParametersComponent    : public juce::Component
{
public:
    DSPParametersComponent (const std::vector<DSPParameterBase*>& demoParams)
    {
        parameters = demoParams;

        for (auto demoParameter : parameters)
        {
            addAndMakeVisible (demoParameter->getComponent());

            auto* paramLabel = new juce::Label ({}, demoParameter->name);

            paramLabel->attachToComponent (demoParameter->getComponent(), true);
            paramLabel->setJustificationType (juce::Justification::centredLeft);
            addAndMakeVisible (paramLabel);
            labels.add (paramLabel);
        }
    }

    void resized() override
    {
        auto bounds = getLocalBounds();
        bounds.removeFromLeft (100);

        for (auto* p : parameters)
        {
            auto* comp = p->getComponent();

            comp->setSize (juce::jmin (bounds.getWidth(), p->getPreferredWidth()), p->getPreferredHeight());

            auto compBounds = bounds.removeFromTop (p->getPreferredHeight());
            comp->setCentrePosition (compBounds.getCentre());
        }
    }

    int getHeightNeeded()
    {
        auto height = 0;

        for (auto* p : parameters)
            height += p->getPreferredHeight();

        return height + 10;
    }

private:
    std::vector<DSPParameterBase*> parameters;
    juce::OwnedArray<juce::Label> labels;
};

//==============================================================================
class AudioFileReaderComponent     : public juce::Component,
                                     private juce::TimeSliceThread,
                                     private juce::Value::Listener,
                                     private juce::ChangeListener
{
public:
    AudioFileReaderComponent()
        : TimeSliceThread ("Audio File Reader Thread"),
          header (audioDeviceManager, formatManager, *this)
    {
        loopState.addListener (this);

        formatManager.registerBasicFormats();
        audioDeviceManager.addAudioCallback (&audioSourcePlayer);
        audioDeviceManager.initialiseWithDefaultDevices (0, 2);

        init();
        startThread();

        setOpaque (true);

        addAndMakeVisible (header);

        setSize (800, 250);
    }

    ~AudioFileReaderComponent() override
    {
        signalThreadShouldExit();
        stop();
        audioDeviceManager.removeAudioCallback (&audioSourcePlayer);
        waitForThreadToExit (10000);
    }

    void paint (juce::Graphics& g) override
    {
        g.setColour (getLookAndFeel().findColour (juce::ResizableWindow::backgroundColourId));
        g.fillRect (getLocalBounds());
    }

    void resized() override
    {
        auto r = getLocalBounds();

        header.setBounds (r.removeFromTop (80));

        r.removeFromTop (20);

        if (parametersComponent.get() != nullptr)
            parametersComponent->setBounds (r.removeFromTop (parametersComponent->getHeightNeeded()).reduced (20, 0));
    }

    //==============================================================================
    bool loadFile (const juce::File& fileToPlay)
    {
        stop();

        audioSourcePlayer.setSource (nullptr);
        getThumbnailComponent().setTransportSource (nullptr);
        transportSource.reset();
        readerSource.reset();

        reader.reset (formatManager.createReaderFor (fileToPlay));

        if (reader.get() != nullptr)
        {
            readerSource.reset (new juce::AudioFormatReaderSource (reader.get(), false));
            readerSource->setLooping (loopState.getValue());

            init();

            return true;
        }

        return false;
    }

    void togglePlay()
    {
        if (playState.getValue())
            stop();
        else
            play();
    }

    void stop()
    {
        playState = false;

        if (transportSource.get() != nullptr)
        {
            transportSource->stop();
            transportSource->setPosition (0);
        }
    }

    void init()
    {
        if (transportSource.get() == nullptr)
        {
            transportSource.reset (new juce::AudioTransportSource());
            transportSource->addChangeListener (this);

            if (readerSource.get() != nullptr)
            {
                if (auto* device = audioDeviceManager.getCurrentAudioDevice())
                {
                    transportSource->setSource (readerSource.get(), juce::roundToInt (device->getCurrentSampleRate()), this, reader->sampleRate);

                    getThumbnailComponent().setTransportSource (transportSource.get());
                }
            }
        }

        audioSourcePlayer.setSource (nullptr);
        currentProcessor.reset();

        if (currentProcessor.get() == nullptr)
            currentProcessor.reset (new DSPProcessor (*transportSource));

        audioSourcePlayer.setSource (currentProcessor.get());

        initParameters();
    }

    void play()
    {
        if (readerSource.get() == nullptr)
            return;

        if (transportSource->getCurrentPosition() >= transportSource->getLengthInSeconds()
             || transportSource->getCurrentPosition() < 0)
            transportSource->setPosition (0);

        transportSource->start();
        playState = true;
    }

    void setLooping (bool shouldLoop)
    {
        if (readerSource.get() != nullptr)
            readerSource->setLooping (shouldLoop);
    }

    AudioThumbnailComponent& getThumbnailComponent()    { return header.thumbnailComp; }

    void initParameters()
    {
        auto& parameters = currentProcessor->getParameters();

        if (parametersComponent.get() != nullptr)
            parametersComponent.reset();

        if (parameters.size() > 0)
        {
            parametersComponent.reset (new DSPParametersComponent (parameters));
            addAndMakeVisible (parametersComponent.get());
        }

        resized();
    }

private:
    //==============================================================================
    class AudioPlayerHeader     : public Component,
                                  private ChangeListener,
                                  private juce::Value::Listener
    {
    public:
        AudioPlayerHeader (juce::AudioDeviceManager& adm,
                           juce::AudioFormatManager& afm,
                           AudioFileReaderComponent& afr)
            : thumbnailComp (adm, afm),
              audioFileReader (afr)
        {
            setOpaque (true);

            addAndMakeVisible (loadButton);
            addAndMakeVisible (playButton);
            addAndMakeVisible (loopButton);

            loadButton.onClick = [this] { openFile(); };
            playButton.onClick = [this] { audioFileReader.togglePlay(); };

            addAndMakeVisible (thumbnailComp);
            thumbnailComp.addChangeListener (this);

            audioFileReader.playState.addListener (this);
            loopButton.getToggleStateValue().referTo (audioFileReader.loopState);
        }

        ~AudioPlayerHeader() override
        {
            audioFileReader.playState.removeListener (this);
        }

        void paint (juce::Graphics& g) override
        {
            g.setColour (getLookAndFeel().findColour (juce::ResizableWindow::backgroundColourId).darker());
            g.fillRect (getLocalBounds());
        }

        void resized() override
        {
            auto bounds = getLocalBounds();

            auto buttonBounds = bounds.removeFromLeft (juce::jmin (250, bounds.getWidth() / 4));
            auto top = buttonBounds.removeFromTop (40);

            loadButton.setBounds (top.removeFromLeft (buttonBounds.getWidth() / 2).reduced (10, 10));
            playButton.setBounds (top.reduced (10, 10));

            loopButton.setSize (0, 25);
            loopButton.changeWidthToFitText();
            loopButton.setCentrePosition (buttonBounds.getCentre());

            thumbnailComp.setBounds (bounds);
        }

        AudioThumbnailComponent thumbnailComp;

    private:
        //==============================================================================
        void openFile()
        {
            audioFileReader.stop();

            chooser = std::make_unique<juce::FileChooser> ("Select an audio file...",
                                                           juce::File{},
                                                           "*.wav;*.mp3;*.aif;");
            auto chooserFlags = juce::FileBrowserComponent::openMode
                              | juce::FileBrowserComponent::canSelectFiles;

            chooser->launchAsync (chooserFlags, [this] (const juce::FileChooser& fc)
            {
                auto f = fc.getResult();

                if (f == juce::File{})
                    return;

                if (! audioFileReader.loadFile (f))
                    juce::NativeMessageBox::showOkCancelBox (juce::AlertWindow::WarningIcon, "Error loading file", "Unable to load audio file", nullptr, nullptr);
                else
                    thumbnailComp.setCurrentFile (f);
            });
        }

        void changeListenerCallback (juce::ChangeBroadcaster*) override
        {
            if (audioFileReader.playState.getValue())
                audioFileReader.stop();

            audioFileReader.loadFile (thumbnailComp.getCurrentFile());
        }

        void valueChanged (juce::Value& v) override
        {
            playButton.setButtonText (v.getValue() ? "Stop" : "Play");
        }

        //==============================================================================
        juce::TextButton loadButton { "Load File..." }, playButton { "Play" };
        juce::ToggleButton loopButton { "Loop File" };

        std::unique_ptr<juce::FileChooser> chooser;

        AudioFileReaderComponent& audioFileReader;
    };

    //==============================================================================
    void valueChanged (juce::Value& v) override
    {
        if (readerSource.get() != nullptr)
            readerSource->setLooping (v.getValue());
    }

    void changeListenerCallback (juce::ChangeBroadcaster*) override
    {
        if (playState.getValue() && ! transportSource->isPlaying())
            stop();
    }

    //==============================================================================
    juce::AudioDeviceManager audioDeviceManager;
    juce::AudioFormatManager formatManager;
    juce::Value playState { juce::var (false) };
    juce::Value loopState { juce::var (false) };

    std::unique_ptr<juce::AudioFormatReader> reader;
    std::unique_ptr<juce::AudioFormatReaderSource> readerSource;
    std::unique_ptr<juce::AudioTransportSource> transportSource;
    std::unique_ptr<DSPProcessor> currentProcessor;

    juce::AudioSourcePlayer audioSourcePlayer;

    AudioPlayerHeader header;

    juce::AudioBuffer<float> fileReadBuffer;

    std::unique_ptr<DSPParametersComponent> parametersComponent;
};
